# Code based on the Pyramid Vision Transformer
# https://github.com/whai362/PVT
# Licensed under the Apache License, Version 2.0
"""
Train and eval functions used in main.py
"""
import csv

import math
import sys
from typing import Iterable, Optional

import torch

from timm.data import Mixup
from timm.utils import accuracy, ModelEma
from sklearn import metrics

from losses import DistillationLoss
import utils


def train_one_epoch(model: torch.nn.Module, criterion: DistillationLoss,
                    data_loader: Iterable, optimizer: torch.optim.Optimizer,
                    device: torch.device, epoch: int, loss_scaler, max_norm: float = 0,
                    model_ema: Optional[ModelEma] = None, mixup_fn: Optional[Mixup] = None,
                    set_training_mode=True,
                    fp32=False):
    model.train()
    metric_logger = utils.MetricLogger(delimiter="  ")
    metric_logger.add_meter('lr', utils.SmoothedValue(window_size=1, fmt='{value:.6f}'))
    header = 'Epoch: [{}]'.format(epoch)
    print_freq = 30

    for samples, targets, word_id, word_mask, topic_id, topic_mask in metric_logger.log_every(data_loader, print_freq, header):
        samples = samples.to(device, non_blocking=True)
        targets = targets.to(device, non_blocking=True)
        word_id = word_id.to(device, non_blocking=True)
        word_mask = word_mask.to(device, non_blocking=True)
        topic_id = topic_id.to(device, non_blocking=True)
        topic_mask = topic_mask.to(device, non_blocking=True)

        if mixup_fn is not None:
            samples, targets = mixup_fn(samples, targets)

        with torch.cuda.amp.autocast(enabled=not fp32):
            outputs = model(samples, word_id, word_mask, topic_id, topic_mask)

            loss = criterion(samples, outputs, targets)

        loss_value = loss.item()

        if not math.isfinite(loss_value):
            print("Loss is {}, stopping training".format(loss_value))
            sys.exit(1)

        optimizer.zero_grad()

        # this attribute is added by timm on one optimizer (adahessian)
        is_second_order = hasattr(optimizer, 'is_second_order') and optimizer.is_second_order
        loss_scaler(loss, optimizer, clip_grad=max_norm,
                    parameters=model.parameters(), create_graph=is_second_order)

        torch.cuda.synchronize()
        if model_ema is not None:
            model_ema.update(model)

        metric_logger.update(loss=loss_value)
        metric_logger.update(lr=optimizer.param_groups[0]["lr"])
    # gather the stats from all processes
    metric_logger.synchronize_between_processes()
    print("Averaged stats:", metric_logger)
    return {k: meter.global_avg for k, meter in metric_logger.meters.items()}


@torch.no_grad()
def evaluate(data_loader, model, device):
    criterion = torch.nn.CrossEntropyLoss()

    metric_logger = utils.MetricLogger(delimiter="  ")
    header = 'Test:'

    # switch to evaluation mode
    model.eval()
    y_pred = []
    y_true = []

    for images, target, word_id, word_mask, topic_id, topic_mask in metric_logger.log_every(data_loader, 10, header):
        y_true.extend(target.detach().tolist())
        images = images.to(device, non_blocking=True)
        target = target.to(device, non_blocking=True)
        with open('/home/user/xxx/MultiModal/TGCA_PVT/tsne/S2E_TGCA_ser_l_100.csv', mode='a', newline='') as file:
            writer = csv.writer(file)
            # 将隐藏层向量转为 numpy 并逐行写入
            labels = target.detach().cpu().numpy()  # [16, 1]
            for label in labels:
                writer.writerow([label])
        word_id = word_id.to(device, non_blocking=True)
        word_mask = word_mask.to(device, non_blocking=True)
        topic_id = topic_id.to(device, non_blocking=True)
        topic_mask = topic_mask.to(device, non_blocking=True)
        # compute output
        with torch.cuda.amp.autocast():
            output = model(images, word_id, word_mask, topic_id, topic_mask)
            loss = criterion(output, target)
        with open('/home/user/xxx/MultiModal/TGCA_PVT/tsne/S2E_TGCA_ser_v_100.csv', mode='a', newline='') as file:
            writer = csv.writer(file)
            hidden_output_np = output.detach().cpu().numpy()  # [16, 6]
            writer.writerows(hidden_output_np)
        probability = torch.max(output, dim=1)[1].data.cpu().numpy()
        y_pred.extend(probability)

        acc1, acc5 = accuracy(output, target, topk=(1, 5))

        batch_size = images.shape[0]
        metric_logger.update(loss=loss.item())
        metric_logger.meters['acc1'].update(acc1.item(), n=batch_size)
        metric_logger.meters['acc5'].update(acc5.item(), n=batch_size)
    
    acc = metrics.accuracy_score(y_true, y_pred)
    print("accuracy=", acc)
    confusion_matrix = metrics.confusion_matrix(y_true, y_pred)
    print("confusion_matrix=", confusion_matrix)
    print(metrics.classification_report(y_true, y_pred, digits=6))
    
    
    # gather the stats from all processes
    metric_logger.synchronize_between_processes()
    print('* Acc@1 {top1.global_avg:.3f} Acc@5 {top5.global_avg:.3f} loss {losses.global_avg:.3f}'
          .format(top1=metric_logger.acc1, top5=metric_logger.acc5, losses=metric_logger.loss))

    return {k: meter.global_avg for k, meter in metric_logger.meters.items()}
